import java.io.*;
import java.util.*;

interface Action {
	void doAction(World world);
}

interface Strategy {
	void doSomething();
}

interface Condition {
	boolean isTrue();
}

class MoveOnCircleStrategy implements Strategy {
	private World world;
	private int frames;
	
	public final static float R = 0.5f;
	public final static float CX = 0.5f;
	public final static float CY = 0.5f;
	public final static float EPS = 1e-5f;
	
	public MoveOnCircleStrategy(World world, int frames) {
		this.world = world;
		this.frames = frames;
	}
	
	private void moveToBorder() {
		System.err.println("b");
		int framesLimit = world.player.Time + frames;
		CPlayer player = world.player;
		float border = R * R;
		
		while (world.player.Time < framesLimit) {
			float dist = BotUtilities.squaredDistance(
					player.PositionX,
					player.PositionY,
					CX,
					CY);
			if (Math.abs(dist - border) < EPS + world.player.Radius)
				break;
			else if (dist > border)
				world.run(CX - player.PositionX, CY- player.PositionY);
			else
				world.run(player.PositionX - CX, player.PositionY - CY);
			world.makeMove();
		}
		System.err.println("e");
	}
	
	public void doSomething() {
		int framesLimit = world.player.Time + frames;
		moveToBorder();
		while (world.player.Time < framesLimit) {
			System.err.println("here");
			CMonster nearest = BotUtilities.getNearestMonster(world);
			if (nearest != null) {
				System.err.println("nearest X : " + nearest.PositionX);
				System.err.println("nearest Y : " + nearest.PositionY);
				BotUtilities.shootAtDirection(world, world.player, nearest);
			}
			float mx = -(world.player.PositionY - 0.5f);
			float my = +(world.player.PositionX - 0.5f);				
			world.run(mx, my);
			world.makeMove();
		}
	}
}

class InCircleCondition implements Condition {
	private World world;
	private float border;
	private float offsetRadius;
	
	public InCircleCondition(World world, float border, float offsetRadius) {
		this.world = world;
		this.border = border;
		this.offsetRadius = offsetRadius;
	}
	
	public boolean isTrue() {
		return CircleConditions.inCircle(world, border, offsetRadius);
	}
}

class OutOfCircleCondition implements Condition {
	private World world;
	private float border;
	private float offsetRadius;
	
	public OutOfCircleCondition(World world, float border, float offsetRadius) {
		this.world = world;
		this.border = border;
		this.offsetRadius = offsetRadius;
	}
	
	public boolean isTrue() {
		return !CircleConditions.inCircle(world, border, offsetRadius);
	}
}

class CircleConditions {
	static boolean[] was = new boolean[4];
	
	static boolean inCircle(World world, float border, float x, float y) {
		Arrays.fill(was, false);
		border = border * border;
		CPlayer player = world.player;
			
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0) {
				float squaredDistance = BotUtilities.squaredDistance(player, monster);
				if (squaredDistance <= border) {
					boolean isRight = monster.PositionX > x;
					boolean isUpper = monster.PositionY > y;
					
					int mask = 0;
					if (isRight)
						mask = mask | 1;
					if (isUpper)
						mask = mask | 2;
					was[mask] = true;
				}
			}
		for (boolean w : was)
			if (!w)
				return false;
		return true;	
	}
	
	static boolean inCircle(World world, float border, float offsetRadius) {
		Arrays.fill(was, false);
		border = border * border;
		CPlayer player = world.player;
		
		float[] c = BotUtilities.getCenterOfMasses(world);
		float cx = c[0];
		float cy = c[1];
		float px = player.PositionX - cx;
		float py = player.PositionY - cy;
		float len = (float) Math.hypot(px, py);
		float off = offsetRadius * (world.player.Radius + BotUtilities.averageMonsterRadius(world));
		float x1 = cx + (px * (len + off) / len);
		float y1 = cy + (py * (len + off) / len);
		
		float x2 = cx + (px * (len - off) / len);
		float y2 = cy + (py * (len - off) / len);
		
		return inCircle(world, border, x1, y1) &&
				inCircle(world, border, x2, y2) &&
		 	   inCircle(world, border, player.PositionX, player.PositionY);
	}
}

class BreakCircleStrategy implements Strategy {
	private World world;
	private int frames, n;
	private Condition[] conditions;
	
	public BreakCircleStrategy(World world, int frames, int n, Condition ... conditions) {
		this.world = world;
		this.frames = frames;
		this.n = n;
		this.conditions = conditions;
	}
	
	private float getBestDirection(int n, int[] density) {
		int best = 0;
		for (int i = 1; i < n; ++i)
			if (density[i] < density[best])
				best = i;
		float phi = (float) (Math.PI * 2 / n);
		return (best + 0.5f) * phi;
	}
	
	public void doSomething() {
		int lastFrame = world.player.Time + frames;
		int[] density = BotUtilities.getMonsterDensity(world, n);
		float direction = getBestDirection(n, density);
		float dxc = (float) Math.cos(direction);
		float dyc = (float) Math.sin(direction);
		
		float alpha = (float) Math.PI / n;
		Point c = new Point(dxc, dyc);
		Point leftP  = c.rotate(-alpha);
		Point rightP = c.rotate(+alpha);

		float lx = leftP.x, ly = leftP.y;
		float rx = rightP.x, ry = rightP.y;

		mainLoop:
		while (world.player.Time < lastFrame) {
			CMonster bestInside = null;
			CMonster bestOutside = null;
			float insideDist = Float.POSITIVE_INFINITY;
			float outsideDist = Float.POSITIVE_INFINITY;

			for (CMonster monster : world.monsters) {
				if (monster.ClassName <= 0)
					continue;
				float squaredDist = BotUtilities.squaredDistance(
						monster.PositionX,
						monster.PositionY,
						world.player.PositionX,
						world.player.PositionY);
				
				float dx = monster.PositionX - world.player.PositionX;
				float dy = monster.PositionY - world.player.PositionY;
				
				boolean inside = 
						BotUtilities.sign(BotUtilities.cross(dx, dy, lx, ly)) <= 0 && 
						BotUtilities.sign(BotUtilities.cross(dx, dy, rx, ry)) >= 0;

				if (inside) {
					if (squaredDist < insideDist) {
						insideDist = squaredDist;
						bestInside = monster;
					}	
				}
				else {
					if (squaredDist < outsideDist) {
						outsideDist = squaredDist;
						bestOutside = monster;
					}	
				}				
			}
			
			if (bestInside != null)
				BotUtilities.shootAtDirection(world, world.player, bestInside);
			else if (bestOutside != null)
				BotUtilities.shootAtDirection(world, world.player, bestOutside);
				
			world.run(dxc, dyc);
			world.makeMove();
			
			
			for (Condition condition : conditions)
				if (condition.isTrue())
					break mainLoop;
		}
	}
}

class DefendStrategy implements Strategy {
	private World world;
	private int frames;
	private Condition[] conditions;
	
	public DefendStrategy(World world, int frames, Condition ... conditions) {
		this.world = world;
		this.frames = frames;
		this.conditions = conditions;
	}
	
	public void doSomething() {
		int initFrame = world.player.Time;
		int lastFrame = initFrame + frames;
		mainLoop: while (world.player.Time < lastFrame) {
			CMonster nearestMonster = BotUtilities.getNearestMonster(world);
			if (nearestMonster != null) {
				BotUtilities.shootAtDirection(world, world.player, nearestMonster);
				float dx = world.player.PositionX - nearestMonster.PositionX;
				float dy = world.player.PositionY - nearestMonster.PositionY;
				world.run(dx, dy);
				world.makeMove();
			}
			for (Condition condition : conditions)
				if (condition.isTrue())
					break mainLoop;
		}
	}
}

class MoveCircle implements Strategy {
	private World world;
	private int frames;
	private Condition[] conditions;
	
	public MoveCircle(World world, int frames, Condition ... conditions) {
		this.world = world;
		this.frames = frames;
		this.conditions = conditions;
	}
	
	public void doSomething() {
		int initFrame = world.player.Time;
		int lastFrame = initFrame + frames;
		mainLoop: while (world.player.Time < lastFrame) {
			CMonster nearestMonster = null;
			float nearestSquaredDist = Float.POSITIVE_INFINITY;
			for (CMonster monster : world.monsters) {
				if (monster.ClassName <= 0)
					continue;
				float squaredDist = BotUtilities.squaredDistance(
						monster.PositionX,
						monster.PositionY,
						world.player.PositionX,
						world.player.PositionY);
				if (squaredDist < nearestSquaredDist) {
					nearestSquaredDist = squaredDist;
					nearestMonster = monster;
				}
			}
			if (nearestMonster != null) {
				BotUtilities.shootAtDirection(world, world.player, BotUtilities.getMostDamagedMonster(world));
				//float[] cm = BotUtilities.getCenterOfMasses(world);				
				float mx = -(world.player.PositionY - 0.5f);
				float my = +(world.player.PositionX - 0.5f);				
				world.run(mx, my);
				world.makeMove();
			}
			for (Condition condition : conditions)
				if (condition.isTrue())
					break mainLoop;
		}
	}
}

public class Task {
	final static float kCircleRadius = 0.4f;
	
	public static void main(String[] args) throws IOException{
		//Check for enough parameters
		if (args.length < 2){
			System.err.println("parameters: IP port");
			System.exit(10);
		}

		//Create the world data (and establish connection with the server)
		World world = new World(args[0], Integer.parseInt(args[1]));
		world.makeMove();
		Strategy strategy = new DefendStrategy(world, 1000000, new InCircleCondition(world, kCircleRadius, 4.5f));
		System.out.println("Exiting defend strategy");
		strategy.doSomething();
		while (true) {
			if (world.player.HitPoints < 0.5f) {
				System.out.println("GO CIRCLE STRATEGY");
				strategy = new MoveOnCircleStrategy(world, 1000000);
				strategy.doSomething();
				System.out.println("exiting CIRCLE STRATEGY");
			}
			strategy = new BreakCircleStrategy(world, 1000000000, 20, new OutOfCircleCondition(world, kCircleRadius, 0.5f));
			strategy.doSomething();
			System.out.println("Exiting break circle strategy");
			strategy = new MoveCircle(world, 1000000000, new InCircleCondition(world, kCircleRadius, 4.5f));
			strategy.doSomething();
			System.out.println("Exiting move circle strategy");
		}
/*		float alpha = (float)0.0;

		//Infinite loop
		while (true){
			//Send our turn and receive the update of the game world
			world.makeMove();

			//Print some information
			System.out.println("frame : " + world.player.Time + " HP: "+ world.player.HitPoints + " x: " + world.player.PositionX + " y: " + world.player.PositionY); 

			alpha += 0.02;
			//Run in circles (set direction)
			world.run((float)Math.cos(alpha), (float)Math.sin(alpha)); 

			//Choose the first monster that is still alive
			for ( int j = 0; j < World.MAXMONSTERS; j++) 
				if (world.monsters[j].ClassName > 0.){
				//Shoot the bastard! (Shoot in the direction of the monster)
					world.shoot(world.monsters[j].PositionX - world.player.PositionX, world.monsters[j].PositionY - world.player.PositionY); 
					break;
				}
		}
*/
	}
}


class Point {
	float x, y;
	Point(float x, float y) {
		this.x = x;
		this.y = y;
	}
	
	Point rotate(float alpha) {
		float c = (float) Math.cos(alpha);
		float s = (float) Math.sin(alpha);
		float nx = x * c - y * s;
		float ny = x * s + y * c;
		return new Point(nx, ny);
	}
}

class BotUtilities {
	public final static float EPS = 1e-5f;
	
	static CMonster getMostDamagedMonster(World world) {
		CMonster result = null;
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0 && (result == null ||
					monster.HitPoints < result.HitPoints))
				result = monster;
		return result;
	}
	
	static CMonster getNearestMonster(World world) {
		CMonster nearestMonster = null;
		float nearestSquaredDist = Float.POSITIVE_INFINITY;
		for (CMonster monster : world.monsters) {
			if (monster.ClassName <= 0)
				continue;
			float squaredDist = BotUtilities.squaredDistance(
					monster.PositionX,
					monster.PositionY,
					world.player.PositionX,
					world.player.PositionY);
			if (squaredDist < nearestSquaredDist) {
				nearestSquaredDist = squaredDist;
				nearestMonster = monster;
			}
		}
		return nearestMonster;
	}
	
	static float squaredDistance(float x, float y, float u, float v) {
		float dx = x - u, dy = y - v;
		return dx * dx + dy * dy;
	}
	
	static float squaredDistance(CPlayer player, CMonster monster) {
		return squaredDistance(player.PositionX, player.PositionY,
				monster.PositionX, monster.PositionY);
	}
	
	static float cross(float x1, float y1, float x2, float y2) {
		return x1 * y2 - x2 * y1;
	}

	static float dot(float x1, float y1, float x2, float y2) {
		return x1 * x2 + y1 * y2;
	}
	
	static float sign(float x) {
		if (Math.abs(x) < EPS)
			return 0;
		return x < 0 ? -1 : 1;
	}

	static void shootAtDirection(World world, CPlayer player, CMonster monster) {
		float dx = monster.PositionX - player.PositionX;
		float dy = monster.PositionY - player.PositionY;
		world.shoot(dx, dy);
	}
	
	static boolean shootAtMonster(World world, CPlayer player, CMonster monster) {
		float dx = monster.PositionX - player.PositionX;
		float dy = monster.PositionY - player.PositionY;
		float vp = (float) Math.hypot(player.VelocityX, player.VelocityY);
		float vm = (float) Math.hypot(monster.VelocityX, monster.VelocityY);
		float A = dx*dx + vp*vp * dy*dy;
		if (Math.abs(A) < EPS) {
			world.shoot(1,1);
			return true;
		}
		
		float B = -2*vm*dx*dy;
		float C = (vm*vm - vp*vp) * dy*dy;
		float D = B*B - 4*A*C;
		if (D < -EPS)
			return false;
		
		float sqrtD = (float) Math.sqrt(D + EPS);
		float x1 = (-B - sqrtD) / (2*A); 
		float x2 = (-B + sqrtD) / (2*A); 
		
		float ang = (float) Math.asin(x1);
		world.shoot((float)Math.cos(ang), (float)Math.sin(ang));
		return true;
	}
	
	static public float[] getCenterOfMasses(World world) {
		float sx = .0f, sy = .0f;
		int totalMonsters = 0;
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0) {
				++totalMonsters;
				sx += monster.PositionX;
				sy += monster.PositionY;
			}
		sx /= totalMonsters;
		sy /= totalMonsters;
		return new float[] { sx, sy };
	}
	
	static float averageMonsterRadius(World world) {
		float res = 0;
		int am = 0;
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0) {
				res += monster.Radius;
				++am;
			}
		return res / am;
	}
	
	static public int[] getMonsterDensity(World world, int n) {
		int[] result = new int[n];
		
		float[] values = new float[World.MAXMONSTERS];
		int totalMonsters = 0;
		float playerX = world.player.PositionX;
		float playerY = world.player.PositionY;
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0) {
				float monsterX = monster.PositionX;
				float monsterY = monster.PositionY;
				float angle = (float) Math.atan2(monsterY - playerY, monsterX - playerX);
				if (angle < 0.)
					angle += Math.PI * 2;
				values[totalMonsters++] = angle;
			}
		values = Arrays.copyOf(values, totalMonsters);
		Arrays.sort(values);
		float dphi = (float) (2 * Math.PI / n), phi = dphi;
		int curIndex = 0;
		for(float v : values) {
			while (curIndex < result.length && phi + EPS < v) {
				++curIndex;
				phi += dphi;
			}
			if (curIndex < result.length)
				++result[curIndex];
		}
		return result;
	}
}
