#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <set>
#include <map>
#include <queue>

#define mp make_pair
#define pb push_back
#define sz(v) ((int)(v).size())
#define all(v) (v).begin(),(v).end() 

using namespace std;

typedef pair<int, int> ii;
typedef long long int64;
typedef long long ll;
typedef vector<int> vi;

template<class T> T abs(T x) { return x > 0 ? x : -x; }
template<class T> T sqr(T x) { return x * x; }

const double inf = 1E100;
const double eps = 1e-6;


typedef double D;

map<int, int> MonsterShoot;

struct P
{
	D x, y;
	P(){}
	P(D x, D y) : x(x), y(y){}
	P operator-(P p)
	{
		return P(x - p.x, y - p.y);
	}
	P operator+(P p)
	{
		return P(x + p.x, y + p.y);
	}
	D len()
	{
		return sqrt(x * x + y * y);
	}
	P operator*(D l)
	{
		return P(x * l, y * l);
	}
	D operator&(P p)
	{
		return x * p.x + y * p.y;
	}
	D operator*(P p)
	{
		return x * p.y - y * p.x;
	}
};

struct L
{
	D a, b, c;
	L(){}
	L(P p, P q)
	{
		a = (p - q).y;
		b = (q - p).x;
		c = -a * p.x - b * p.y;
	}
	D operator() (P p)
	{
		return a * p.x + b * p.y + c;
	}
};

const int BULLETS_PER_MONSTER = 0;
const int CHANCE_OF_RESET = 30;
const double FIRING_DANGER = 0.6;
const double ANGLE_SHIFT = 0.1;
const double DPS_LIMIT = 0.07;
const double GALTING_GUN = 0.05;
const double DIST_RESET = 0.07 * 0.07;

const double SIDE_CHARGE = 57;
const double BOT_CHARGE = 1.0;
const double BULLET_CHARGE = 0.3;
const double FORCE_EPS = 0.001;
const double CIRCULATION_COEFF = 0.1;
const double DIRECTION_STABILITY = 57;

const double ANGLE_CHARGE = 20.0;

double EvaluateDanger(CMonster monster)
{
	double res = 1.0;
	double bite_per_sec = monster.BiteDamage / max<double>(1E-6, monster.TeethCooldownPeriod);
	double dps = monster.BulletDamage / max<double>(1E-6, monster.GunCooldownPeriod);
	double velocity = monster.BulletVelocity;
	if (abs(dps) > eps)
	{
		//cerr << dps << "\n";
		res *= FIRING_DANGER;
	}
	if (abs(dps) > DPS_LIMIT)
		res *= GALTING_GUN;
	return res;
}

double EvaluateMonster(World *world, int monsterIndex)
{
	if (world->monsters[monsterIndex].ClassName < 0)
	{
		MonsterShoot[monsterIndex] = 0;
		return inf;
	}
	double x0 = world->player.PositionX;
	double y0 = world->player.PositionY;
	double x = world->monsters[monsterIndex].PositionX;
	double y = world->monsters[monsterIndex].PositionY;
	double dist = (x - x0) * (x - x0) + (y - y0) * (y - y0);
	int ratio = (int)((world->monsters[monsterIndex].MaxHitPoints + world->player.BulletDamage - eps) / world->player.BulletDamage);
	if (MonsterShoot[monsterIndex] > BULLETS_PER_MONSTER + ratio)
	{
		if (dist < DIST_RESET)
			MonsterShoot[monsterIndex] = 0;
		else if (rand() % CHANCE_OF_RESET == 0)
			MonsterShoot[monsterIndex] = 0;
		return inf;
	}
	
	P ourDir(world->player.VelocityX, world->player.VelocityY);
	P toTarget(x - x0, y - y0);
	double angle = abs(atan2(ourDir * toTarget, ourDir & toTarget));
	double res = dist * (angle + ANGLE_SHIFT) * sqrt(eps + world->monsters[monsterIndex].HitPoints);
	res *= EvaluateDanger(world->monsters[monsterIndex]);
	return res;
}


void ShootMonster(World *world, int index)
{
	double mx = world->monsters[index].PositionX - world->player.PositionX;
	double my = world->monsters[index].PositionY - world->player.PositionY;
	double vx = world->monsters[index].VelocityX;
	double vy = world->monsters[index].VelocityY;
	double V = world->player.BulletVelocity;
	double A = sqr(vx) + sqr(vy) - sqr(V);
	double B = 2 * (mx * vx + my * vy);
	double C = sqr(mx) + sqr(my);
	double discr = sqr(B) - 4 * A * C;
	if (discr < -eps) {//No way
		return;
	}
	if (discr < 0) discr = 0;
	double t1 = (-B - sqrt(discr)) / 2 / A;
	double t2 = (-B + sqrt(discr)) / 2 / A;
	if (t1 < 0 && t2 < 0) {//No way
		return;
	}
	if (t1 < 0) t1 = t2;
	if (t2 < 0) t2 = t1;
	double t = min(t1, t2);
	MonsterShoot[index]++;
	world->shoot(mx + vx * t, my + vy * t);
}


void AddForce(double& fx, double& fy, double sx, double sy, double x, double y, double charge, double decrement = 0.0, bool useForceEps = false)
{
	double d2 = sqr(x - sx) + sqr(y - sy);
	d2 = sqr(max<double>(0.0, sqrt(d2) - decrement));
	if (useForceEps)
		d2 = max(d2, FORCE_EPS);
	fx += (sx - x) / d2 * charge;
	fy += (sy - y) / d2 * charge;
}

void AddCirculation(double &fx, double &fy, double x, double y)
{
	x -= 0.5;
	y -= 0.5;
	fx += y * CIRCULATION_COEFF;
	fx += -x * CIRCULATION_COEFF;
}

void Go(World *world)
{
	double fx = world->player.VelocityX, fy = world->player.VelocityY;
	double tmp = sqrt(fx * fx + fy * fy);
	if (tmp < eps) tmp = eps;
	fx /= tmp;
	fy /= tmp;
	fx *= DIRECTION_STABILITY;
	fy *= DIRECTION_STABILITY;
	double sx = world->player.PositionX;
	double sy = world->player.PositionY;

	// monsters
	for (int i = 0; i < world->MAXMONSTERS; ++i)
		if (world->monsters[i].ClassName >= 0)
		{
			AddForce(fx, fy, sx, sy, world->monsters[i].PositionX, world->monsters[i].PositionY, BOT_CHARGE, world->monsters[i].Radius + world->player.Radius, true);
		}
	
	// sides
	AddForce(fx, fy, sx, sy, sx, 0.0, SIDE_CHARGE);
	AddForce(fx, fy, sx, sy, sx, 1.0, SIDE_CHARGE);
	AddForce(fx, fy, sx, sy, 0.0, sy, SIDE_CHARGE);
	AddForce(fx, fy, sx, sy, 1.0, sy, SIDE_CHARGE);

	// Angles
	AddForce(fx, fy, sx, sy, 0.0, 0.0, ANGLE_CHARGE);
	AddForce(fx, fy, sx, sy, 0.0, 1.0, ANGLE_CHARGE);
	AddForce(fx, fy, sx, sy, 1.0, 0.0, ANGLE_CHARGE);
	AddForce(fx, fy, sx, sy, 1.0, 1.0, ANGLE_CHARGE);

	// center
	//AddForce(fx, fy, sx, sy, 0.5, 0.5, CENTER_CHARGE);

	// matrix
	for (int i = 0; i < world->MAXMISSILES; ++i)
	{
		if (world->missiles[i].ClassName == 200 || world->missiles[i].ClassName == -1) continue;
		AddForce(fx, fy, sx, sy, world->missiles[i].PositionX, world->missiles[i].PositionY, BOT_CHARGE, world->player.Radius, true);
		/*if ((we - missile).len() > BULLET_LEN_LIMIT) continue;
		P missile(world->missiles[i].PositionX, world->missiles[i].PositionY);
		P we(world->player.PositionX, world->player.PositionY);
		P dir(world->missiles[i].VelocityX, world->missiles[i].VelocityY);
		L l(missile, missile + dir);
		double dist = l(we) / (sqr(l.a) + sqr(l.b));
		P perp(l.a, l.b);
		P chargePos = we - perp * dist;
		AddForce(fx, fy, sx, sy, chargePos.x, chargePos.y, BULLET_CHARGE / (we - missile).len());*/
	}

	// circulation
	AddCirculation(fx, fy, sx, sy);
	world->run(fx, fy);
}

int main(int argc, char** argv) {
	//freopen("error.log", "w", stderr);
	srand(0xdead);
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		Go(world);

		double optVal = inf;
		int opt = -1;
		for ( int j = 0; j < world->MAXMONSTERS; j++) {
			double cur = EvaluateMonster(world, j);
			if (optVal > cur)
				optVal = cur, opt = j;
		}
		if (opt != -1)
			if (world->monsters[opt].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				ShootMonster(world, opt);
			}
	}

	return 0;
}
