#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <complex>
#include <map>
#include <vector>
#include <algorithm>

using namespace std;

typedef complex<double> cd;

double eps = 1e-8;
double MAX = 10e10;
const double PI = acos(-1.0);
World *world;

double scal(cd & a, cd & b) {
	return real(a * conj(b));
}

double vec(cd & a, cd & b) {
	return imag(conj(a) * b);
}

cd dirToShoot(CMonster & m, CPlayer & p) {
	double vb = p.BulletVelocity;
	cd vm(m.VelocityX, m.VelocityY);
	cd vp(p.VelocityX, p.VelocityY);
	cd rm(m.PositionX, m.PositionY);
	cd rp(p.PositionX, p.PositionY);
	double s = scal(vm, rm - rp);
	double t;
	double a = norm(vm) - vb*vb;
	if (abs(a) < eps) {
		t = norm(rm - rp) / (2*s);
	} else {
		double discr = (s*s - norm(rm - rp) * a);
		if (discr < 0) return cd(0,0);
		double t1 = (-s + sqrt(discr)) / a;
		double t2 = (-s - sqrt(discr)) / a;
		if (min(t1, t2) > 0) {
			t = min(t1, t2);
		} else if (max(t1, t2) < 0) {
			return cd(0,0);
		} else 
			t = max(t1, t2);
	}
	cd monstpos = rm + t * vm;
	if (min(monstpos.real(), monstpos.imag()) < 0 || max(monstpos.real(), monstpos.imag()) > 1)
		return cd(0,0);
	return monstpos - rp;
}

bool isInter(CMonster & m, CMissile & b) {
	cd vm(m.VelocityX, m.VelocityY);
	cd vb(b.VelocityX, b.VelocityY);
	cd rm(m.PositionX, m.PositionY);
	cd rb(b.PositionX, b.PositionY);
	if (abs(vm - vb) < eps) return false;
	double sc = scal(rm - rb, vb - vm) / abs(vb - vm);
	double d = sqrt(norm(rm - rb) - sc * sc);
	if (d < m.Radius && sc > 0) return true;
	return false;
}

double countIncomBullets(CMonster & m, World * world) {
	double bulhp = 0;
	for (int i = 0; i < world->MAXMISSILES; ++i) {
		if (world->missiles[i].ClassName > 0 ) {
			if (isInter(m, world->missiles[i])) bulhp += world->missiles[i].Damage;
		}
	}
	return bulhp;
}

int needMoreBullets(CMonster & m, World * world) {
	double inbul = countIncomBullets(m, world);
	double todie = m.HitPoints - inbul;
	if (todie <= 0) return 0;
	return int(ceil(todie / world->player.BulletDamage));
}

double timeToBite(CMonster & m, CPlayer & p) {
	if (m.BiteDamage <= eps) return MAX;
	double vb = m.MaxVelocity;
	cd vp(m.VelocityX, m.VelocityY);
	cd vm(p.VelocityX, p.VelocityY);
	cd rp(m.PositionX, m.PositionY);
	cd rm(p.PositionX, p.PositionY);
	double s = scal(vm, rm - rp);
	double t;
	double a = norm(vm) - vb*vb;
	if (abs(a) < eps) {
		t = norm(rm - rp) / (2*s);
	} else {
		double discr = (s*s - norm(rm - rp) * a);
		if (discr < 0) return MAX;
		double t1 = (-s + sqrt(discr)) / a;
		double t2 = (-s - sqrt(discr)) / a;
		if (min(t1, t2) > 0) {
			t = min(t1, t2);
		} else if (max(t1, t2) < 0) {
			return MAX;
		} else 
			t = max(t1, t2);
	}
	if (t < 0) return MAX;
	return t;
}

const int NUM_OPTIONS = 100;
const double ANGLE_STEP = 2 * PI / NUM_OPTIONS;
const cd STEP_MULTIPLIER(exp(cd(0., 1.) * ANGLE_STEP));

inline double WeightForChaseTime(double time) {
	return 1.0 / time;
}

inline cd GetRunningDirection(CPlayer &p) {
	cd playerCoords(p.PositionX, p.PositionY);
	cd center(0.5, 0.5);
	double dist = abs(playerCoords - center);
	const double R = 0.44;
	if (abs(dist - R) < 1e-2) {
		return (playerCoords - center) * cd(0., 1.);
	} else {
		if (dist < 1e-4)
			return cd(1.0, 0.0);
		else {
			if (dist < R)
				return playerCoords - center;
			else
				return - (playerCoords - center);
		}
	}
	/*cd curDirection(1., 0.);
	for (int i = 0; i < NUM_OPTIONS; i++, curDirection *= STEP_MULTIPLIER) {
		double curWeight = 0.0;
		for (int j = 0; j < World::MAXMONSTERS; j++) {
		}
	}*/
}

double monstPlus(CMonster & m, CPlayer & p) {
	cd vm(m.VelocityX, m.VelocityY);
	cd vp(p.VelocityX, p.VelocityY);
	cd rm(m.PositionX, m.PositionY);
	cd rp(p.PositionX, p.PositionY);
	double dist = abs(rm - rp);
	double time = timeToBite(m, p);
	double threshold = 2.;
	double shootThreshold = 0.7;
	if (time < threshold) return time;
	else {
		double ans = threshold + dist;
		if (m.BulletDamage > eps) {
			return ans;
		} else {
			return ans + shootThreshold;
		}
	}
}

typedef pair<double, int> pdi;


int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;

	//Infinite loop
	
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		alpha += 0.012;
		//Run in circles (set direction)
		cd direction = GetRunningDirection(world->player);
		world->run(real(direction),imag(direction));

		//Choose the first monster that is still alive
		vector<pdi> timebite;
		for (int curmon = 0 ; curmon < world->MAXMONSTERS; ++curmon) { //;  curmon = (curmon + 1) % world->MAXMONSTERS) {
			if (world->monsters[curmon].ClassName >= 0 &&
					world->monsters[curmon].HitPoints > 0 &&
					needMoreBullets(world->monsters[curmon], world) > 0) {
						cd shootdir = dirToShoot(world->monsters[curmon], world->player);
						if (abs(shootdir) < eps) continue;
						timebite.push_back(
							pdi(monstPlus(world->monsters[curmon], world->player),
								curmon));
				//world->shoot(shootdir.real(),shootdir.imag());
				//break;
			}
		}
		sort(timebite.begin(), timebite.end());
		if (timebite.size() > 0) {
			cd shootdir = dirToShoot(world->monsters[timebite.front().second],
				                     world->player);
			world->shoot(shootdir.real(),shootdir.imag());
		}
	}
	return 0;
}
