#include "GeomObjects.h"

double2 RandomVector() {
	double2 res;
	do {
		res.x = 2*trandom() - 1;
		res.y = 2*trandom() - 1;
	} while ((res ^ res) > 1.0);
	return res;
}

double2 RandomDirection() {
	double2 res;
	do {
		res = RandomVector();
	} while (!res < EPS);
	return ForceVectorNorm(res, 1.0);
}

double2 LimitVectorNorm(double2 vect, double maxnorm) {
	double l = !vect;
	if (l > maxnorm) return vect * (maxnorm / l);
	else return vect;
}

double2 ForceVectorNorm(double2 vect, double norm) {
	double l = !vect;
	if (l > EPS) return vect * (norm / l);
	else return vect;
}

bool InField(double2 pos) {
	return pos.x >= 0.0 && pos.x <= FIELD_SIZE && 
		   pos.y >= 0.0 && pos.y <= FIELD_SIZE;
}

void ForceToField(double2 &pos, double radius) {
	double l = radius;
	double r = FIELD_SIZE - radius;
	pos = std::max(pos, double2(l, l));
	pos = std::min(pos, double2(r, r));
}

std::vector<int2> GetGridCells(const MovingCircle &circle) {
	int2 li, ri;
	for (int u = 0; u<2; u++) {
		//get bbox
		double cmin = std::min(circle.st[u], circle.end[u]);
		double cmax = std::max(circle.st[u], circle.end[u]);
		double r = circle.rad;
		cmin -= r;  cmax += r;

		//scale it to grid coordinates
		cmin /= FIELD_SIZE;
		cmax /= FIELD_SIZE;
		cmin = (cmin + GRID_OUTSIDE_WIDTH) / (1.0 + 2*GRID_OUTSIDE_WIDTH);
		cmax = (cmax + GRID_OUTSIDE_WIDTH) / (1.0 + 2*GRID_OUTSIDE_WIDTH);

		//get indices
		static const int GS = GRID_SIZE;
		cmin *= GS;
		cmax *= GS;
		int imin = int(floor(cmin));
		int imax = int( ceil(cmax));

		//cut outside of world
		if (imin < 0) imin = 0;
		if (imax < 1) imax = 1;
		if (imin > GS-1) imin = GS-1;
		if (imax > GS) imax = GS;

		li[u] = imin;  ri[u] = imax;
	}

	std::vector<int2> res((ri.x-li.x)*(ri.y-li.y));
	int k = 0;
	for (int x = li.x; x<ri.x; x++)
		for (int y = li.y; y<ri.y; y++)
			res[k++] = int2(x, y);

	return res;
}

double IntersectionTime(const MovingCircle &a, const MovingCircle &b) {
	double r = a.rad + b.rad;
	double2 s = b.st - a.st;
	double2 v = (b.end-b.st) - (a.end-a.st);

	double qc = (s ^ s) - (r * r);
	if (qc < 0.0) return 0.0;
	double qa = v ^ v;
	if (qa < EPS) return INF;
	double qb = s ^ v;

	double qd = qb * qb - qa * qc;
	if (qd < 0.0) return INF;
	qd = sqrt(qd);
	double t1 = (-qb - qd) / qa;
	double t2 = (-qb + qd) / qa;

	if (t1 < 0.0 || t1 > 1.0) t1 = INF;
	if (t2 < 0.0 || t2 > 1.0) t2 = INF;
	return std::min(t1, t2);
}

bool HasIntersection(const MovingCircle &a, const MovingCircle &b) {
	return IntersectionTime(a, b) < INF/2;
}

double2 ShootMovingTarget(double2 pos, double2 vel, double speed) {
	double qa = (vel ^ vel) - (speed * speed);
	double qb = (pos ^ vel);
	double qc = (pos ^ pos);
	if (qa > -EPS) return ForceVectorNorm(pos, speed);
	double qd = qb*qb - qa*qc;
	qd = sqrt(qd);
	double t1 = (-qb - qd) / qa;
	double t2 = (-qb + qd) / qa;
	if (t1 < 0.0) t1 = INF;
	if (t2 < 0.0) t2 = INF;
	double t = std::min(t1, t2);
	return ForceVectorNorm(pos + vel*t, speed);
}
